/***************************************************************************
 *   Copyright (C) 2006 by AST   *
 *   tsyplyaev@gmail.com   *
 *   ICQ: 279-533-134                          *
 *   This is a part of work done by AST.       *
 *   If you want to use it, please contact me. *
 ***************************************************************************/

#pragma once

//**************************************
#include "GLSystem.h"
#include "Font.h"
#include <vector>
//**************************************

namespace ASTEngine {

class Widget;

//---------------------------------------------------------------------------
//Desc: Gui struct. Created one time
//---------------------------------------------------------------------------
class GUI {
public:
	static GUI *create(const String &folder);
	static GUI *get();

	~GUI();

	void addWidget(Widget *widget);
	void removeWidget(Widget *widget);

	void draw();
	
	void setAlpha(float alpha) { this->alpha = alpha; };
	float getAlpha() { return alpha; };

private:
	GLTexture *windowTex;
	
	GLTexture *buttonTex;
	GLTexture *buttonFocusedTex;

	GLTexture *checkBoxTex;
	GLTexture *checkBoxCheckedTex;

	GLTexture *hSliderTex;

	Font *font;
	float alpha;

	std::vector<Widget*> widgets;

	friend class Widget;
	friend class WidgetButton;
	friend class WidgetCheckBox;
	friend class WidgetHSlider;
	friend class WidgetLabel;
	friend class WidgetWindow;
};

//---------------------------------------------------------------------------
//Desc: GUI Widget item
//---------------------------------------------------------------------------
class Widget {
public:
	virtual void addWidget(Widget *widget);
	virtual void removeWidget(Widget *widget);
	
	virtual bool isPressed();
	virtual bool wasPressed();
	virtual bool wasReleased();
	virtual bool isFocused();

	virtual	void draw();
	virtual void checkCallbacks() {};

	virtual void setPosition(int x, int y) { posX = x; posY = y; };
	virtual int getX() { if(getParent()) return posX + parent->getX(); else return posX; };
	virtual int getY() { if(getParent()) return posY + parent->getY(); else return posY; };
	
	virtual void setSize(int w, int h) { width = w; height = h; };
	virtual int getWidth() { return width; };
	virtual int getHeight() { return height; };
	
	virtual Widget *getParent() { return parent; };
	virtual void setParent(Widget *parent) { this->parent = parent; };

	virtual void setActive(bool act) { active = act; };  
	virtual void toggleActive() { active = !active; };
	virtual bool isActive() { return active; };
	
	typedef void (*WidgetCallback)(Widget*);

protected:
	int posX, posY;
	int width, height;
	
	bool pressed;
	bool active;

	Widget *parent;
	std::vector<Widget*> children;

	friend class WidgetRadioGroup;
};

//---------------------------------------------------------------------------
//Desc: GUI button element
//---------------------------------------------------------------------------
class WidgetButton : public Widget {
public:
	WidgetButton(const String &text);
	~WidgetButton();
	virtual	void draw();

private:
	String text;
	int size;
};

//---------------------------------------------------------------------------
//Desc: GUI check box
//---------------------------------------------------------------------------
class WidgetCheckBox : public Widget {
public:
	WidgetCheckBox(const String &text);
	~WidgetCheckBox();
	
	bool isChecked() { return checked; };
	void setChecked(bool checked) { this->checked = checked; };
	void toggleCheck() { checked = !checked; };
	
	virtual	void draw();

private:
	bool checked;

	String text;
};

//---------------------------------------------------------------------------
//Desc: GUI radio group
//---------------------------------------------------------------------------
class WidgetRadioGroup : public Widget {
public:
	WidgetRadioGroup();
	~WidgetRadioGroup();

	void addCheckBox(WidgetCheckBox *cbox);
	void removeCheckBox(WidgetCheckBox *cbox);

	virtual	void draw();

private:
	std::vector<WidgetCheckBox*> checkBoxes;
};

//---------------------------------------------------------------------------
//Desc: GUI label element
//---------------------------------------------------------------------------
class WidgetLabel : public Widget {
public:
	WidgetLabel(const String &text);
	~WidgetLabel();

	virtual void draw();
	
	void setText(const String text) { this->text = text; width = height * text.size(); };

private:
	String text;
};

//---------------------------------------------------------------------------
//GUI horizontal slider element
//---------------------------------------------------------------------------
class WidgetHSlider : public Widget {
public:
	WidgetHSlider();
	~WidgetHSlider();

	virtual void draw();

	int getValue() { return value; };
	void setValue(int v) { value = v; pos = width*v*0.01; };

private:
	int value;
	int pos;
};

//---------------------------------------------------------------------------
//Window GUI element
//---------------------------------------------------------------------------
class WidgetWindow : public Widget {
public:
	WidgetWindow();
	~WidgetWindow();

	virtual void draw();
};

};