/***************************************************************************
 *   Copyright (C) 2006 by AST   *
 *   tsyplyaev@gmail.com   *
 *   ICQ: 279-533-134                          *
 *   This is a part of work done by AST.       *
 *   If you want to use it, please contact me. *
 ***************************************************************************/

//***************************************************************************
#include "PhysSystem.h"
#include "PhysBody.h"
#include "Log.h"
#include "Config.h"
//***************************************************************************

namespace ASTEngine {

PhysSystem *physSystem;

//contact-callback------------------
int PhysSystem::contactBegin(const NewtonMaterial* material, const NewtonBody* body0, const NewtonBody* body1) {
	physSystem->pBody0 = (PhysBody*)NewtonBodyGetUserData(body0);
	physSystem->pBody1 = (PhysBody*)NewtonBodyGetUserData(body1);

	physSystem->impactSpeed = 0;
	
	return 1;
};

//contact-callback------------------
int PhysSystem::contactProcess(const NewtonMaterial* material, const NewtonContact* contact) {
	float speed;

	speed = NewtonMaterialGetContactNormalSpeed(material, contact);
	if(speed > physSystem->impactSpeed) {
		physSystem->impactSpeed = speed;
		NewtonMaterialGetContactPositionAndNormal(material, physSystem->impactPosition, physSystem->impactNormal);
	};

	return 1;	
};	

//contact-callback------------------
void PhysSystem::contactEnd(const NewtonMaterial* material) {
	if(physSystem->impactSpeed > 15) {
		if(physSystem->pBody0) { 
			if(physSystem->pBody0->impactSrc) {
				if(!physSystem->pBody0->impactSrc->isPlaying()) {
					physSystem->pBody0->impactSrc->setPosition(physSystem->impactPosition);
					physSystem->pBody0->impactSrc->play();
				};
			}; 
		};

		if(physSystem->pBody1) { 
			if(physSystem->pBody1->impactSrc) {
				if(!physSystem->pBody1->impactSrc->isPlaying()) {
					physSystem->pBody1->impactSrc->setPosition(physSystem->impactPosition);
					physSystem->pBody1->impactSrc->play();
				};
			}; 
		};
	};
};

//---------------------------------------------------------------------------
//Desc:    creates new PhysSystem
//Params:  -
//Returns: pointer to new PhysSystem
//---------------------------------------------------------------------------
PhysSystem *PhysSystem::create() {
	if(physSystem) { return physSystem; };

	physSystem = new PhysSystem();
	
	Log::write("PhysSystem::create()");
	Log::write("{");
	Log::write("}\n");

	physSystem->nWorld = NewtonCreate(NULL, NULL);
	
	physSystem->worldSize = Vec3(10000, 10000, 10000);
	NewtonSetWorldSize(physSystem->nWorld, -physSystem->worldSize * 0.5, physSystem->worldSize * 0.5);

	physSystem->accTimeSlice = 0.0f;

	physSystem->defaultID = NewtonMaterialGetDefaultGroupID(physSystem->nWorld);

	NewtonMaterialSetDefaultSoftness(physSystem->nWorld, physSystem->defaultID, physSystem->defaultID, 0.05f);
	NewtonMaterialSetDefaultElasticity(physSystem->nWorld, physSystem->defaultID, physSystem->defaultID, 0.4f);
	NewtonMaterialSetDefaultCollidable(physSystem->nWorld, physSystem->defaultID, physSystem->defaultID, 1);
	NewtonMaterialSetDefaultFriction(physSystem->nWorld, physSystem->defaultID, physSystem->defaultID, 1.0f, 0.5f);
	NewtonMaterialSetCollisionCallback(physSystem->nWorld, physSystem->defaultID, physSystem->defaultID, NULL, contactBegin, contactProcess, contactEnd); 
		 
	return physSystem;
};

//---------------------------------------------------------------------------
//Desc:    returns existing PhysSystem
//Params:  -
//Returns: pointer to existing PhysSystem
//---------------------------------------------------------------------------
PhysSystem *PhysSystem::get() {
	return physSystem;
};


//---------------------------------------------------------------------------
//Desc:    PhysSystem destructor
//Params:  -
//Returns: -
//---------------------------------------------------------------------------
PhysSystem::~PhysSystem() {
	NewtonDestroy(nWorld);
};

//---------------------------------------------------------------------------
//Desc:    updates PhysSystem engine
//Params:  dTime - time from the last frame in miliseconds
//Returns: -
//---------------------------------------------------------------------------
void PhysSystem::update(float dTime) {
	accTimeSlice += dTime;
	
	while(accTimeSlice > 12.0f) {
		NewtonUpdate(nWorld, (12.0f / 1000.0f));
        accTimeSlice -= 12.0f;
	};

	intersectionParam = 100000.0;
};

//---------------------------------------------------------------------------
//Desc:    sets PhysSystem world size
//Params:  worldSize - world size 
//Returns: -
//---------------------------------------------------------------------------
void PhysSystem::setWorldSize(const Vec3 &worldSize) {
	this->worldSize = worldSize;
	NewtonSetWorldSize(nWorld, -worldSize * 0.5, worldSize * 0.5);
};

//---------------------------------------------------------------------------
//Desc:    Newton callback
//Params:  -
//Returns: -
//---------------------------------------------------------------------------
float PhysSystem::rayCastFilter(const NewtonBody* body, const float* normal, int collisionID, void* userData, float iParam) {
	if(iParam < physSystem->intersectionParam) {
		physSystem->intersectionParam = iParam;
		physSystem->intersectionNormal = Vec3(normal[0],normal[1],normal[2]);
		physSystem->intersectedBody = (PhysBody*)NewtonBodyGetUserData(body);
	};
	return iParam;
};

//---------------------------------------------------------------------------
//Desc:    sets PhysSystems world by ray
//Params:  src - ray src, dst 
//Returns: -
//---------------------------------------------------------------------------
PhysBody *PhysSystem::intersectWorldByRay(const Vec3 &src, const Vec3 &dst, Vec3 &normal, Vec3 &point) {
	NewtonWorldRayCast(nWorld, src, dst, rayCastFilter, NULL, NULL);
	
	if(intersectedBody) {
		point = src + (dst - src) * intersectionParam;
		normal = intersectionNormal;
		return intersectedBody;
	};
	return NULL;
};

};










